from stdlib._string cimport string_view, hash_string_view
from libc.string cimport strlen, strncpy
from libc.stdlib cimport malloc, free


cdef cypclass Str:
    string_view _s

    __init__(self, const char *s):
        cdef size_t l = strlen(s)
        cdef char *_s = <char *> malloc(l + 1)
        strncpy(_s, s, l + 1)
        self._s = string_view(_s, l)

    __dealloc__(self):
       cdef char *s = <char *> self._s.data()
       self._s = string_view()
       free(s)

    size_t __len__(self):
        return self._s.size()

    bint __eq__(self, Str other):
        return self._s == other._s

    size_t __hash__(self):
        return hash_string_view()(self._s)

    char __getitem__(self, int index) except 0:
        cdef int end = self._s.size()
        cdef int idx = index
        if index < 0:
            index = -index
            idx = end - index
        if index >= end:
            with gil:
                raise ValueError('index out of range')
        return self._s.data()[idx]

    int find(self, Str s, size_t start=0, size_t stop=0):
        if start < stop and stop <= self._s.size():
            sw = string_view(self._s.data(), stop)
            return sw.find(s._s, start)
        return self._s.find(s._s, start)

    Str slice(self, int start=0, int stop=0) except NULL:
        cdef int end = self._s.size()
        cdef int tmp = stop
        if stop <= 0:
            stop = -stop
            tmp = end - stop
        if stop > end:
            with gil:
                raise ValueError('slice bounds out of range')
        stop = tmp
        if start < 0:
            start = -start
            if start > end:
                with gil:
                    raise ValueError('slice bounds out of range')
            start = end - start
        if start >= stop:
            with gil:
                raise ValueError('slice bounds out of order')
        cdef size_t size = stop - start
        cdef char *s = <char *> malloc(size + 1)
        strncpy(s, self._s.data() + start, size)
        s[size] = 0
        result = Str()
        result._s = string_view(s, size)
        return result

    Str __add__(self, Str other):
        cdef size_t l1 = self._s.size()
        cdef size_t l2 = other._s.size()
        cdef Str cat = new Str()
        cdef char *s = <char *> malloc(l1 + l2 + 1)
        strncpy(s, self._s.data(), l1)
        strncpy(s + l1, other._s.data(), l2 + 1)
        cat._s = string_view(s, l1 + l2)
        return cat

    @staticmethod
    const char * to_c_str(Str s):
        if s is NULL:
            return NULL
        return s._s.data()

    @staticmethod
    Str steal_c_str(const char *s, size_t size):
        result = Str()
        result._s = string_view(s, size)
        return result

